@@ -18,6 +18,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <gdk/gdkkeysyms.h>
+#include <unistd.h>
 
 #include <sys/time.h>
 #include <time.h>
@@ -72,11 +73,17 @@ GtkWidget *menu_item_view_fullscreen = NULL;
 GtkWidget *view[3] = { NULL, NULL, NULL };
 GtkTextBuffer *buffer[3] = { NULL, NULL, NULL };
 
-int key_event = 0;
+typedef struct key_event_s {
+  int key;
+  struct key_event_s *next;
+} key_event_t;
+
+key_event_t *key_event = NULL;
 
 static void screen_refresh(void) { printf("screen_refresh()\n"); } 
 static void screen_redraw(void) { printf("screen_redraw()\n"); } 
 
+#if 0
 static GtkWidget *gtk_label_small_new(char *name) {
   GtkWidget *label = gtk_label_new("");
   char *markup = g_markup_printf_escaped("<span size='x-small'>%s</span>", name);
@@ -90,6 +97,7 @@ static void gtk_label_small_set_text(GtkWidget *label, char *name) {
   gtk_label_set_markup(GTK_LABEL(label), markup);
   g_free(markup);
 }
+#endif
 
 struct {
   int code;
@@ -430,12 +438,6 @@ static void draw_attdef(const char *type, int from, int to) {
 
 static void draw_gmsg_input(const char *s, int pos) { 
   printf("draw_gmsg_input(%s, %d)\n", s, pos);
-
-  if(!gmsg_input_entry) return;
-
-  gtk_widget_show(gmsg_input_entry);
-  gtk_entry_set_text(GTK_ENTRY(gmsg_input_entry), s);
-  gtk_editable_set_position(GTK_EDITABLE(gmsg_input_entry), pos);
 }
 
 static void clear_gmsg_input(void) { 
@@ -444,25 +446,36 @@ static void clear_gmsg_input(void) {
   /* remove entire input line if present */
   if(!gmsg_input_entry) return;
 
-  gtk_entry_set_text(GTK_ENTRY(partyline_input_entry), "");
+  gtk_entry_set_text(GTK_ENTRY(gmsg_input_entry), "");
   gtk_widget_hide(gmsg_input_entry);
 }
 
 static void draw_partyline_input(const char *s, int pos) { 
   printf("draw_partyline_input(%s, %d)\n", s, pos); 
 
-  if(!partyline_input_entry) return;
-
-  /* */
-  gdk_window_focus(gtk_widget_get_parent_window(partyline_input_entry), 0);
-
-  gtk_entry_set_text(GTK_ENTRY(partyline_input_entry), s);
-  gtk_editable_set_position(GTK_EDITABLE(partyline_input_entry), pos);
+  if(!strlen(s)) 
+    gtk_entry_set_text(GTK_ENTRY(partyline_input_entry), "");
 }
 
-
 static void setup_winlist(void) { printf("setup_winlist()\n"); }
 
+static void key_enqueue(int key) {
+  key_event_t **keyP = &key_event;
+
+  /* search end of key queue */
+  while(*keyP)
+    keyP = &(*keyP)->next;
+
+  /* attach new entry */
+  *keyP = g_new0(key_event_t,1);
+  (*keyP)->key = key;  
+}
+
+static void key_enqueue_string(char *str) {
+  while(*str)
+    key_enqueue(*str++);
+}
+
 static gboolean playfield_clicked_event(GtkWidget *widget, GdkEventButton *event,
 				      gpointer user_data) {
   int i = (int)g_object_get_data(G_OBJECT(widget), "index");
@@ -472,7 +485,7 @@ static gboolean playfield_clicked_event(GtkWidget *widget, GdkEventButton *event
     printf("my_playernum = %d\n", my_playernum);
     printf("clicked %d, player %d\n", i, FIELD2PLAYER(i));
 
-    key_event = '0' + FIELD2PLAYER(i);
+    key_enqueue('0' + FIELD2PLAYER(i));
   }
 
   return FALSE;
@@ -516,13 +529,11 @@ GtkWidget *playfield_new(int i) {
   gtk_signal_connect(GTK_OBJECT(playfield[i]), "button_press_event",
 		     (GtkSignalFunc)playfield_clicked_event, NULL);
 
-  Field *f = &fields[FIELD2PLAYER(i)-1];
-
   /* TODO: should be taken from game date? */
   int x, y;
   for (y = 0; y < 22; y++) 
     for (x = 0; x < 12; x++) 
-      field_state[i][y][x] = (*f)[y][x];
+      field_state[i][y][x] = 0;
 
   return playfield[i];
 }
@@ -542,7 +553,7 @@ static gboolean special_queue_clicked_event(GtkWidget *widget, GdkEventButton *e
   /* issue "delete special" command if not entering text */
   if(!GTK_WIDGET_VISIBLE(gmsg_input_entry)) {
     printf("clicked specials\n");
-    key_event = 'd';
+    key_enqueue('d');
   }
 
   return FALSE;
@@ -706,31 +717,21 @@ void on_window_destroy (GtkWidget *widget, gpointer data) {
   window = NULL;
 
   /* send F10 (quit) to tetrinet engine */
-  key_event = K_F10;
+  key_enqueue(K_F10);
 }
 
 static int transtab[][2] = {
-  { GDK_F1, K_F1 },
-  { GDK_F2, K_F2 },
-  //  { GDK_F3, K_F3 },   // winlist (nyi)
-  { GDK_F10, K_F10 },
-
-  { GDK_Up, K_UP },
-  { GDK_Down, K_DOWN },
-  { GDK_Left, K_LEFT },
+  { GDK_Up,    K_UP    },
+  { GDK_Down,  K_DOWN  },
+  { GDK_Left,  K_LEFT  },
   { GDK_Right, K_RIGHT },
-
-  { GDK_KP_Enter, '\r' },    // n810 return key is in fact keypad enter
+  { ' ',       ' '     },    // block drop
 
 #ifdef USE_HILDON
   { GDK_Return, ' ' },       // direction pad center is return key -> 
                              // map to space for block drop
-#else
-  { GDK_Return, '\r' },
 #endif
 
-  { GDK_BackSpace, '\b' },
-
   { 0, 0 }
 };
 
@@ -738,22 +739,19 @@ gboolean on_window_key_press(GtkWidget *widget,
 			 GdkEventKey *event, gpointer data) {
   int handled = FALSE;
 
-  printf("key event %d\n", event->keyval);
+  /* map game keys only in field mode */
+  if(gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook)) == view_idx[FIELDS]) {
 
-  if(event->keyval >= 32 && event->keyval <= 127)  {
-    key_event = event->keyval;
-    return TRUE;
-  }
-
-  int i = 0;
-  while(event->keyval != transtab[i][0] && transtab[i][0])
-    i++;
-
-  if(transtab[i][0]) {
-    printf("translated key\n");
-
-    key_event = transtab[i][1];
-    return TRUE;
+    int i = 0;
+    while(event->keyval != transtab[i][0] && transtab[i][0])
+      i++;
+  
+    if(transtab[i][0]) {
+      printf("translated key\n");
+      
+      key_enqueue(transtab[i][1]);
+      return TRUE;
+    }
   }
 
   switch(event->keyval) {
@@ -797,14 +795,23 @@ gboolean on_window_key_press(GtkWidget *widget,
 /* this is ugly and generates 100% cpu load. this really needs to */
 /* be done more clever */
 static int wait_for_input(int msec) { 
+  int state = 0;
 
   printf("wait_for_input(%d)\n", msec); 
 
+  if(key_event) {
+    printf("key %d in queue\n", key_event->key);
+
+    key_event_t *cur = key_event;
+    state = key_event->key;
+    key_event = key_event->next;
+    g_free(cur);
+    return state;
+  }
+
   //  if(msec > 0) { while(1) gtk_main_iteration(); }
   //  g_assert(msec < 0);
 
-  int state = 0;
-
   struct timeval start, end;
   if(msec > 0)
     gettimeofday(&start, NULL);
@@ -827,15 +834,22 @@ static int wait_for_input(int msec) {
       state = -1;
     }
 
-    while(gtk_events_pending()) {
-      gtk_main_iteration();
-      if(key_event) {
-	state = key_event;
-	key_event = 0;
-	//	printf("sending key %d (%c)\n", state, state); 
-	return state;
+    if(gtk_events_pending()) {
+      while(gtk_events_pending()) {
+	gtk_main_iteration();
+	
+	if(key_event) {
+	  key_event_t *cur = key_event;
+	  state = key_event->key;
+	  key_event = key_event->next;
+	  
+	  g_free(cur);
+	  //	printf("sending key %d (%c)\n", state, state); 
+	  return state;
+	}
       }
-    }
+    } else
+      usleep(1000);
 
     if(msec > 0) {
       gettimeofday(&end, NULL);
@@ -872,7 +886,7 @@ GtkWidget *text_window_new(int bufnum) {
   gtk_text_buffer_create_mark(buffer[bufnum], "end", &end, FALSE);
 
   /* Change default font throughout the widget */
-  PangoFontDescription *font_desc = pango_font_description_from_string("Courier 9");
+  PangoFontDescription *font_desc = pango_font_description_from_string("Courier 10");
   gtk_widget_modify_font(view[bufnum], font_desc);
   pango_font_description_free(font_desc);
 
@@ -888,7 +902,8 @@ static void on_notebook_page_change(GtkNotebook     *notebook,
   printf("switched to page %d\n", page_num);
 
   int keys[] = { K_F1, K_F2, K_F3 };
-  key_event = keys[page_num];
+  printf("enqueue key %d\n", keys[page_num]);
+  key_enqueue(keys[page_num]);
 }
 
 static void screen_setup(void) { 
@@ -911,28 +926,28 @@ static void setup_partyline(void) {
 
 Interface tty_interface = {
 
-    wait_for_input,
-
-    screen_setup,
-    screen_refresh,
-    screen_redraw,
-
-    draw_text,
-    clear_text,
-
-    setup_fields,
-    draw_own_field,
-    draw_other_field,
-    draw_status,
-    draw_specials,
-    draw_attdef,
-    draw_gmsg_input,
-    clear_gmsg_input,
-
-    setup_partyline,
-    draw_partyline_input,
-
-    setup_winlist
+  wait_for_input,
+  
+  screen_setup,
+  screen_refresh,
+  screen_redraw,
+  
+  draw_text,
+  clear_text,
+  
+  setup_fields,
+  draw_own_field,
+  draw_other_field,
+  draw_status,
+  draw_specials,
+  draw_attdef,
+  draw_gmsg_input,
+  clear_gmsg_input,
+  
+  setup_partyline,
+  draw_partyline_input,
+  
+  setup_winlist
 };
 
 static gboolean gmsg_window_clicked_event(GtkWidget *widget, GdkEventButton *event,
@@ -941,7 +956,12 @@ static gboolean gmsg_window_clicked_event(GtkWidget *widget, GdkEventButton *eve
   /* issue "start enter text" command if not already entering text */
   if(!GTK_WIDGET_VISIBLE(gmsg_input_entry)) {
     printf("clicked gmsg window\n");
-    key_event = 't';
+
+    if(gmsg_input_entry) {
+      gtk_widget_show(gmsg_input_entry);
+      gtk_widget_grab_focus(gmsg_input_entry);
+      return TRUE;
+    }
   }
 
   return FALSE;
@@ -1043,6 +1063,25 @@ int tfprintf(FILE *stream, const char *format, ...) {
   return 0;
 }
 
+void on_partyline_activate(GtkEntry *entry, gpointer  user_data) {
+  char *msg = (char*)gtk_entry_get_text(entry);
+
+  printf("partyline activate with message \"%s\"\n", msg);
+
+  key_enqueue_string(msg);
+  key_enqueue('\r');
+}
+
+void on_gmsg_activate(GtkEntry *entry, gpointer  user_data) {
+  char *msg = (char*)gtk_entry_get_text(entry);
+
+  printf("gmsg activate with message \"%s\"\n", msg);
+
+  key_enqueue('t');
+  key_enqueue_string(msg);
+  key_enqueue('\r');
+}
+
 /* this is a pretty ugly hack but it allows us to preempt the main() in tetrinet.c */
 /* we interfere with normal startup to do three things: */
 /* 1. forward the command line parameters to gtk */
@@ -1101,8 +1140,6 @@ int main(int ac, char **av) {
   blocks_pix = blocks_load();
 
   notebook = gtk_notebook_new();
-  g_signal_connect(G_OBJECT(notebook), "switch-page", 
-		   G_CALLBACK(on_notebook_page_change), NULL);
 
   /* +++++++++++++++++++++ create "field" view ++++++++++++++++++++++ */
 
@@ -1117,6 +1154,9 @@ int main(int ac, char **av) {
   GtkWidget *ivbox = gtk_vbox_new(0, FALSE);
   gtk_box_pack_start(GTK_BOX(ivbox), text_window_new(1), TRUE, TRUE, 0);
   gmsg_input_entry = gtk_entry_new();
+  gtk_signal_connect(GTK_OBJECT(gmsg_input_entry), "activate",
+		     (GtkSignalFunc)on_gmsg_activate, NULL);
+  
   gtk_box_pack_start(GTK_BOX(ivbox), gmsg_input_entry, FALSE, FALSE, 0);
 
   gtk_signal_connect(GTK_OBJECT(view[1]), "button_press_event",
@@ -1167,6 +1207,8 @@ int main(int ac, char **av) {
   gtk_box_pack_start_defaults(GTK_BOX(ivbox), text_window_new(0));
 
   partyline_input_entry = gtk_entry_new();
+  gtk_signal_connect(GTK_OBJECT(partyline_input_entry), "activate",
+		     (GtkSignalFunc)on_partyline_activate, NULL);
   gtk_box_pack_start(GTK_BOX(ivbox), partyline_input_entry, 0,0,0);
 
   view_idx[PARTYLINE] = gtk_notebook_append_page(GTK_NOTEBOOK(notebook), ivbox, 
@@ -1178,6 +1220,9 @@ int main(int ac, char **av) {
   gtk_widget_show_all(GTK_WIDGET(window));
   gtk_widget_hide(gmsg_input_entry);
 
+  g_signal_connect(G_OBJECT(notebook), "switch-page", 
+		   G_CALLBACK(on_notebook_page_change), NULL);
+
   /* let gtk do its thing */
   while(gtk_events_pending())
     gtk_main_iteration();
